import xml.etree.ElementTree as ET
import app
import PythonQt

import traceback
import requests

class ChemblWebServices:
    def __init__( self ):
        self.url_start = 'https://www.ebi.ac.uk/chembl/api/data/'
    
    'https://www.ebi.ac.uk/chembl/api/data/similarity/CN1C(=O)C=C(c2cccc(Cl)c2)c3cc(ccc13)[C@@](N)(c4ccc(Cl)cc4)c5cncn5C/80' 
    
    def similaritySearch( self, smiles, threshold ):
        url = self.url_start + 'similarity/' + requests.utils.quote(smiles) + '/' + str(int(100*threshold))
        searcher = requests.get(url)
        return searcher.text
        
"""
ChemblMolecule represents molecule data returned by Chembl web services
"""
class ChemblMolecule:
    def __init__(self):
        self.id = ''
        self.smiles = ''
        self.mostAPkA = 0.0
        self.mostBPkA = 0.0
       
       
def get_child(element):
    # Python version dependent
    method = getattr(element, "getchildren", None)
    if callable(method):
        return element.getchildren()
    else:
        return [child for child in element]
"""
Reads Chembl molecule data from an xml string returned by Chembl webservices.
"""
def fromXML(xmlString):

    molList = []
    
    root = ET.fromstring(xmlString)
    if not root.tag == 'response':
        print(root.tag + ' expected ' +  'response')
        return molList
    
    # Python version dependent
    children = get_child(root)
                
    if len(children) == 0:
        return molList
        
    molecules = children[0]            
    if not molecules.tag == 'molecules':
        return molList

    for mol in get_child(molecules):
        molecule = ChemblMolecule()
        molecule.id = mol.find('molecule_chembl_id').text
        structures = mol.find('molecule_structures')
        if structures is not None:
            molecule.smiles = structures.find('canonical_smiles').text
        properties = mol.find('molecule_properties')
        if properties is not None:
            apka = properties.find('acd_most_apka')
            if apka is not None and apka.text is not None:
                molecule.mostAPkA = float(apka.text)
            bpka = properties.find('acd_most_bpka')
            if bpka is not None and bpka.text is not None:
                molecule.mostBPkA = float(bpka.text)
        molList.append(molecule)            
    
    return molList        
    
def chemblHyperLink(chemblid):
    return ('<a href="https://www.ebi.ac.uk/chembl/compound/inspect/' +
            chemblid +
            '">' +
            chemblid +
            '</a>')
 
def createDataSet(compounds):
    dataset = app.Dataset()
    idSet = set()
    smilesEntries = []
    linkEntries = []
    idEntries = []
    acidicpkas = []
    basicpkas = []
    for compound in compounds:
        if compound.id in idSet:
            continue 
        smilesEntries.append(dataset.molecule(compound.smiles))
        idEntries.append(app.Dataset.string(compound.id))        
        linkEntries.append(app.Dataset.string(chemblHyperLink(compound.id)))   
        if compound.mostAPkA == 0.0:
            acidicpkas.append(dataset.createInvalidEntry(app.cont))
        else:
            acidicpkas.append(dataset.cont(compound.mostAPkA, 0.0))
        if compound.mostBPkA == 0.0:
            basicpkas.append(dataset.createInvalidEntry(app.cont))
        else:
            basicpkas.append(dataset.cont(compound.mostBPkA, 0.0))
        
        idSet.add(compound.id)        
    
    dataset.insertColumn(app.Dataset.meta('Structure', app.mol), smilesEntries, 0)        
    dataset.insertColumn(app.Dataset.meta('ChEMBL IDs', app.string), idEntries, 1)        
    dataset.insertColumn(app.Dataset.meta('ChEMBL links', app.string), linkEntries, 2)        
    dataset.insertColumn(dataset.createNumberHeader('most acidic pKa'), acidicpkas, 3)   
    dataset.insertColumn(dataset.createNumberHeader('most basic pKa'), basicpkas, 4)   
    return dataset.pack()
    
"""
Searches for similar compounds in ChEMBL and reports results in a new data set column
""" 	
def findSimilarChemblCompounds(appui):

    dataset = appui.importDS("")
    indices = appui.getSelectedRows("") 

    ds = app.Dataset()
    ds.parse(dataset)
    
    rowIndices = set(indices)

    # Ask the user for a similarity threshold. 
    threshold = float(appui.query('Similarity', 'Enter a similarity value:\n(Note: If the value you choose is low, you may\nsee errors because too many results are returned)', '0.97'))
    if threshold < 0.0 or threshold > 1.0:
        appui.show("Please specify a similarity value between 0.0 and 1.0")
        return False
    
    columnName = 'Similar ChEMBL compounds (' + str(threshold) + ')'
        
    # Find the structure column and any existing similarity column we may replace
    structureColumn = -1
    chemblColumn = -1
    for colIndex in range(ds.columnSz()):
        type = ds.columnType(colIndex)
        if structureColumn == -1 and type == app.mol:
            structureColumn = colIndex
        if chemblColumn == -1 and type == app.string and ds.columnName(colIndex) == columnName:
            chemblColumn = colIndex
    if structureColumn == -1:
        appui.show("No structures available to use in search")
        return False
        
    progDlg = PythonQt.SimpleProgress()
    progDlg.modal = True
    progDlg.setWindowTitle('Find Similar ChEMBL Compounds')
    progDlg.setMessage('Running ChEMBL similarity search web service')
    progDlg.show()
    
    newEntries = [] 
    allCompounds = []
    gotAtLeastOneResult = False
    for rowIndex in range(ds.rowSz()):
        progress = (100.0 * rowIndex) / ds.rowSz()
        progDlg.setProgress(progress)
        smiles = ds.stringval(structureColumn, rowIndex)
        ids = ''
        if not chemblColumn == -1:
            ids = ds.stringval(chemblColumn, rowIndex)
        if (len(rowIndices) == 0 or rowIndex in rowIndices) and progDlg.visible:
            try:  
                searcher = ChemblWebServices()           
                xml = searcher.similaritySearch(smiles, threshold)
                compounds = fromXML(xml)
                ids = ''
                for compound in compounds:
                    allCompounds.append(compound)
                    ids = ids + chemblHyperLink(compound.id) + ' ' 
                if ids == '':
                    ids = 'None'
                else:
                    gotAtLeastOneResult = True
            except:
                traceback.print_exc()
                ids = 'Error'            
                print("Exception in findSimilarChemblCompounds")        
                
        newEntries.append(app.Dataset.string(ids))

    if chemblColumn == -1:
        chemblColumn = structureColumn + 1
    else:        
        appui.removeColumnFromDataSet("", chemblColumn)

    progDlg.hide()        
        
    # Insert results column in the current data set
    columnHeader = app.Dataset.meta(columnName, app.string)
    ds.insertColumn( columnHeader, newEntries, chemblColumn )
    appui.addColumnToDataSet("", chemblColumn, ds.packColumn(chemblColumn))
    
    # Optionally create a separate data set containing the similar ChEMBL compounds
    if gotAtLeastOneResult:
        w = PythonQt.Warning()
        w.setWindowTitle('Create Data Set')
        w.setMessage('Would you like to create a data set containing the similar ChEMBL compounds?')
        if w.run():
            appui.exportDS(columnName, createDataSet(allCompounds))
        w.delete()
    #if appui.query('Create Data Set', 'Would you like to create a data set containing the similar ChEMBL compounds?', 'yes') == 'yes':
    #    appui.exportDS(columnName, ChemblSimilaritySearch.createDataSet(allCompounds))

    
    return True


        
       