from __future__ import print_function

import app
import PythonQt
import requests
import json

class PDBLigandData:
    def __init__(self):
        self.query = ''
        self.smiles = ''
        self.compoundID = ''
        self.score = ''

class PDBLigandDataSet:
    def __init__(self):
        self.compounds = []
    
    def addLigand(self, ligandData):
        for lig in ligandData:
            newLig = PDBLigandData()
            newLig.query = lig.query
            newLig.smiles = lig.smiles
            newLig.compoundID = lig.compoundID
            newLig.score = lig.score
            self.compounds.append(newLig)
        
        return
    
    def createDataset(self):
        dataset = app.Dataset()
        colQuery = []
        colLigand = []
        colCID = []
        colScore = []
        
        for c in self.compounds:
            colQuery.append(dataset.molecule(c.query))
            colLigand.append(dataset.molecule(c.smiles))
            colCID.append(dataset.string(getHyperLink(c.compoundID)))
            colScore.append(dataset.cont(c.score, 0))
        
        dataset.insertColumn(dataset.meta('PDB Ligand', app.mol), colLigand, 0)
        dataset.insertColumn(dataset.meta('Query Structure', app.mol), colQuery, 1)
        dataset.insertColumn(dataset.meta('Ligand ID', app.string), colCID, 2)
        dataset.insertColumn(dataset.createNumberHeader('Similarity'), colScore, 3)
        
        return dataset.pack()

class PDBWebService:
    def __init__(self):
        self.entityScore = {}
        self.ligandScore = {}
        self.ligandInfo = {}
    
    def searchPDB(self, smiles):
        query = '''
        {{
            "query": {{
                "type": "terminal",
                "service": "chemical",
                "parameters": {{
                    "value": "{}",
                    "type": "descriptor",
                    "descriptor_type": "SMILES",
                    "match_type": "fingerprint-similarity"
                }}
            }},
            "return_type": "non_polymer_entity"
        }}
        '''.format(smiles)
        
        urlbase = 'https://search.rcsb.org/rcsbsearch/v2/query?json='
        urlreq = urlbase + requests.utils.quote(query)
        
        response = requests.get(urlreq)
        
        if response.status_code == 200:
            data = json.loads(response.text)
            for entry in data['result_set']:
                self.entityScore[entry['identifier']] = entry['score']
        elif response.status_code == 204:
            print('Searching PDB: No results!')
            return False
        else:
            print('Searching PDB Error:', resp.getcode())
            return False
        
        return True

    def getEntityData(self):
        query = '''
        {{
            nonpolymer_entities(entity_ids: {}) {{
                rcsb_id,
                pdbx_entity_nonpoly {{
                    comp_id
                }}
            }}
        }}
        '''.format(str([str(e) for e in self.entityScore.keys()]).replace("'", '"'))
        
        urlbase = 'https://data.rcsb.org/graphql?query='
        urlreq = urlbase + requests.utils.quote(query)
        resp = requests.get(urlreq)
        
        data = json.loads(resp.text)
        if data['data'] == None:
            print('Retrieving data Errors:\n', data['errors'])
            return False
        elif len(data['data']['nonpolymer_entities']) == 0:
            print('Retrieving entity data: No results!')
            return False
        
        for item in data['data']['nonpolymer_entities']:
            entid = item['rcsb_id']
            cid = item['pdbx_entity_nonpoly']['comp_id']
            if not cid in self.ligandScore.keys():
                self.ligandScore[cid] = self.entityScore[entid]
        
        return True

    def getLigandData(self):
        query = '''
        {{
            chem_comps(comp_ids: {}) {{
                rcsb_chem_comp_descriptor {{
                    comp_id,
                    SMILES_stereo
                }}
            }}
        }}
        '''.format(str([str(e) for e in self.ligandScore.keys()]).replace("'", '"'))
        
        urlbase = 'https://data.rcsb.org/graphql?query='
        urlreq = urlbase + requests.utils.quote(query)
        resp = requests.get(urlreq)
        
        data = json.loads(resp.text)
        if data['data'] == None:
            print('Retrieving data Errors:\n', data['errors'])
            return False
        elif len(data['data']['chem_comps']) == 0:
            print('Retrieving ligand data: No results!')
            return False
        
        for item in data['data']['chem_comps']:
            cid = item['rcsb_chem_comp_descriptor']['comp_id']
            smi = item['rcsb_chem_comp_descriptor']['SMILES_stereo']
            self.ligandInfo[cid] = smi
        
        return True

    def getSimilarLigand(self, smiles):
        result = self.searchPDB(smiles)
        if not result:
            return None
        
        result = self.getEntityData()
        if not result:
            return None

        result = self.getLigandData()
        if not result:
            return None
        
        ligs = []
        for cid in self.ligandInfo.keys():
            lig = PDBLigandData()
            lig.query = smiles
            lig.smiles = self.ligandInfo[cid]
            lig.compoundID = cid
            lig.score = self.ligandScore[cid]
            ligs.append(lig)
    
        return ligs

def getHyperLink(pdbID):
    return ('<a href="https://www.rcsb.org/ligand/' + pdbID + '">' +
                    pdbID + '</a> ')

def getColumnIndex(dataset, pdbColName):
    # Find the structure column and any existing similarity column we may replace
    structureColumn = -1
    pdbColumn = -1
    for colIndex in range(dataset.columnSz()):
        coltype = dataset.columnType(colIndex)
        if structureColumn == -1 and coltype == app.mol:
            structureColumn = colIndex
        if pdbColumn == -1 and coltype == app.string and \
        dataset.columnName(colIndex) == pdbColName:
            pdbColumn = colIndex
    
    return (structureColumn, pdbColumn)

def findSimilarPDBLigands(appui):
    rows = appui.getSelectedRows('')
    if len(rows) == 0:
        appui.show('Please select one molecule at least!')
        return
    
    dataset = appui.importDS('')
    ds = app.Dataset()
    ds.parse(dataset)
    
    pdbColName = 'Similar PDB ligands'
    structureColumn, pdbColumn = getColumnIndex(ds, pdbColName)

    if structureColumn == -1:
        appui.show('No structures available to use in search')
        return
    
    if pdbColumn == -1:
        newCol = [app.Dataset.string('')] * ds.rowSz()
    else:
        newCol = ds.getColumn(pdbColumn)
    
    progDlg = PythonQt.SimpleProgress()
    progDlg.modal = True
    progDlg.setWindowTitle('Find Similar PDB Ligands')
    progDlg.setMessage('Running PDB similarity search web service')
    progDlg.show()
    
    pdbData = PDBLigandDataSet()
    gotAtLeastOneResult = False
    for n in rows:
        progress = (100.0 * n) / len(rows)
        progDlg.setProgress(progress)
        mol = ds.stringval(structureColumn, n)
        pdbweb = PDBWebService()
        ligands = pdbweb.getSimilarLigand(mol)
        if ligands != None:
            pdbData.addLigand(ligands)
            gotAtLeastOneResult = True
            strcell = ''
            for lig in ligands:
                strcell += getHyperLink(lig.compoundID)
            newCol[n] = app.Dataset.string(strcell)
    progDlg.hide()
    
    if not gotAtLeastOneResult:
        appui.show('No similar structures found!')
        return
    
    if pdbColumn == -1:
        pdbColumn = structureColumn + 1
    else:
        appui.removeColumnFromDataSet('', pdbColumn)
    
    ds.insertColumn(app.Dataset.meta(pdbColName, app.string), 
                                    newCol, pdbColumn)
    appui.addColumnToDataSet('', pdbColumn, ds.packColumn(pdbColumn))
    
    # Create a new data set containing similar PDB compounds
    if gotAtLeastOneResult:
        w = PythonQt.Warning()
        w.setWindowTitle('Create Data Set')
        w.setMessage('Would you like to create a data set containing the similar PDB ligands?')
        if w.run():
            appui.exportDS(pdbColName, pdbData.createDataset())
        w.delete()

    return
