# PubChemSimilaritySearch.py - general component for StarDrop plugin scripts
#
# Copyright (C) Optibrium Ltd 2016
# stardrop-support@optibrium.com
#
# This script uses the web-based services provided by 
# PubChem to look-up information about chemical structures

# For advice or feedback, contact stardrop-support@optibrium.com
import requests
import xml.etree.ElementTree as ET
import app
import PythonQt

"""
PubChemCompoundData represents compound data returned by PubChem web services

Currently only describes PubChem ID and CanonicalSmiles
"""
class PubChemCompoundData:
    def __init__(self):
        """
        Constructor for the PubChemCompoundData class
        """
        self.compoundID = ''
        self.smiles = ''

class PubChemSearcher:
    def __init__( self ): 
        self.dummy = 'dummy'
    
    """
    Uses the PubChem web service to perform a similarity search using specified smiles and threshold
    returning results in XML format
    xmlResults = similaritySearch('OCCCC=O', 0.97)
    """
    def similaritySearch( self, smiles, threshold ):
        url = ('https://pubchem.ncbi.nlm.nih.gov/rest/pug/compound/fastsimilarity_2d/smiles/' + 
              requests.utils.quote(smiles) +
              '/property/CanonicalSmiles/XML?Threshold=' +
              str(int(threshold*100)))
        searcher = requests.get(url)
        return searcher.text    
            
    """
    Convert XML similarity search results to a list of compounds
    xmlResults = similaritySearch('OCCCC=O', 0.97)
    """            
    def readCompounds(self, xmlString):
        compoundList = []
        root = ET.fromstring(xmlString)
        uri = '{http://pubchem.ncbi.nlm.nih.gov/pug_rest}'
        if not root.tag == uri+'PropertyTable':
            print(root.tag + ' expected ' +  'PropertyTable')
            return compoundList
            
        for compound in root.findall(uri+'Properties'):
            compoundData = PubChemCompoundData()
            compoundData.compoundID = compound.find(uri+'CID').text
            compoundData.smiles = compound.find(uri+'CanonicalSMILES').text
            compoundList.append(compoundData)
        
        return compoundList           

def pubChemHyperLink(pubChemID):
    return ('<a href="https://pubchem.ncbi.nlm.nih.gov/compound/' +
            pubChemID +
            '">' +
            pubChemID +
            '</a>')
        
def createDataSet(compounds):
    dataset = app.Dataset()
    idSet = set()
    smilesEntries = []
    linkEntries = []
    idEntries = []
    for compound in compounds:
        if compound.compoundID in idSet:
            continue 
        smilesEntries.append(dataset.molecule(compound.smiles))
        idEntries.append(app.Dataset.string(compound.compoundID))        
        linkEntries.append(app.Dataset.string(pubChemHyperLink(compound.compoundID)))   
        idSet.add(compound.compoundID)        
    
    dataset.insertColumn(app.Dataset.meta('Structure', app.mol), smilesEntries, 0)        
    dataset.insertColumn(app.Dataset.meta('PubChem IDs', app.string), idEntries, 1)        
    dataset.insertColumn(app.Dataset.meta('PubChem links', app.string), linkEntries, 2)        
    return dataset.pack()

"""
Searches for similar compounds in PubChem and reports results in a new data set column
""" 
def findSimilarPubChemCompounds(appui):

    dataset = appui.importDS("")
    indices = appui.getSelectedRows("") 

    ds = app.Dataset()
    ds.parse(dataset)
    
    rowIndices = set(indices)

    # Ask the user for a similarity threshold. 
    threshold = float(appui.query('Similarity', 'Enter a similarity value:\n(Note: If the value you choose is low, you may\nsee errors because too many results are returned)', '0.97'))
    if threshold < 0.0 or threshold > 1.0:
        appui.show("Please specify a similarity value between 0.0 and 1.0")
        return False
    
    columnName = 'Similar PubChem compounds (' + str(threshold) + ')'
        
    # Find the structure column and any existing similarity column we may replace
    structureColumn = -1
    pubChemColumn = -1
    for colIndex in range(ds.columnSz()):
        type = ds.columnType(colIndex)
        if structureColumn == -1 and type == app.mol:
            structureColumn = colIndex
        if pubChemColumn == -1 and type == app.string and ds.columnName(colIndex) == columnName:
            pubChemColumn = colIndex
    if structureColumn == -1:
        appui.show("No structures available to use in search")
        return False
        
    progDlg = PythonQt.SimpleProgress()
    progDlg.modal = True
    progDlg.setWindowTitle('Find Similar PubChem Compounds')
    progDlg.setMessage('Running PubChem similarity search web service')
    progDlg.show()
    
    newEntries = [] 
    allCompounds = []
    gotAtLeastOneResult = False
    for rowIndex in range(ds.rowSz()):
        progress = (100.0 * rowIndex) / ds.rowSz()
        progDlg.setProgress(progress)
        smiles = ds.stringval(structureColumn, rowIndex)
        ids = ''
        if not pubChemColumn == -1:
            ids = ds.stringval(pubChemColumn, rowIndex)
        if (len(rowIndices) == 0 or rowIndex in rowIndices) and progDlg.visible:
            try:  
                searcher = PubChemSearcher()           
                xml = searcher.similaritySearch(smiles, threshold)
                compounds = searcher.readCompounds(xml)
                ids = ''
                for compound in compounds:
                    allCompounds.append(compound)
                    ids = ids + pubChemHyperLink(compound.compoundID) + ' ' 
                if ids == '':
                    ids = 'None'
                else:    
                    gotAtLeastOneResult = True
            except:
                ids = 'Error'             
                print("Exception in findSimilarPubChemCompounds")        
                
        newEntries.append(app.Dataset.string(ids))

    if pubChemColumn == -1:
        pubChemColumn = structureColumn + 1
    else:        
        appui.removeColumnFromDataSet("", pubChemColumn)

    progDlg.hide()        
        
    # Insert results column in the current data set
    columnHeader = app.Dataset.meta(columnName, app.string)
    ds.insertColumn( columnHeader, newEntries, pubChemColumn )
    appui.addColumnToDataSet("", pubChemColumn, ds.packColumn(pubChemColumn))
    
    # Optionally create a separate data set containing the similar PubChem compounds
    if gotAtLeastOneResult:
        w = PythonQt.Warning()
        w.setWindowTitle('Create Data Set')
        w.setMessage('Would you like to create a data set containing the similar PubChem compounds?')
        if w.run():
            appui.exportDS(columnName, createDataSet(allCompounds))
        w.delete()
    #if appui.query('Create Data Set', 'Would you like to create a data set containing the similar PubChem compounds?', 'yes') == 'yes':
    #    appui.exportDS(columnName, PubChemSimilaritySearch.createDataSet(allCompounds))

    
    return True
        
       